import zswag
import sys

from zswag.test.calc import working_dir
import zswag.test.calc.client as client
import zswag.test.calc.server as server
import calculator.api as calculator

mode = sys.argv[1] if len(sys.argv) > 1 else ""
host, port = sys.argv[2].split(':') if len(sys.argv) > 2 and ':' in sys.argv[2] else ("localhost", 5000)

if mode == "client":
    client.run(host, port)
elif mode == "server":
    app = zswag.OAServer(
        controller_module=server,
        service_type=calculator.Calculator.Service,
        yaml_path=working_dir+"/api.yaml",
        zs_pkg_path=working_dir)
    app.run(host=host, port=port)
elif mode == "path":
    print(working_dir)
else:
    print("Usage: python3 -m calc {server|client} [host:port]")



from enum import Enum
from typing import List, Dict
import re
import glob
import os

RPC_DOC_PATTERN = r"""
        service\s+{service_name}\s+\{{   # service MyService {{ (double-braces due to later .format())
            (?:\n|.)*                    #   ...
            /\*\*?\s*                    #   /**
                ((?:[^*]|\*[^/])*)       #     (doc-string) -> captured
            \*/                          #   */
            \s*([A-Za-z0-9_]*)\s+        #   (return-type) -> captured
            {rpc_name}                   #   method-name
                \s*\(\s*                 #   (
                    ([A-Za-z0-9_]*)      #     (argument-type) -> captured
                \s*\)                    #   )
"""

STRUCT_PATTERN = r"""
          /\*\!                          # /*!
            ((?:[^!]|![^*]|!\*[^/])*)    #   (doc-string) -> captured
          !\*/\s+                        # !*/
          struct\s+{name}                # struct NAME
"""

SERVICE_PATTERN = r"""
          /\*\!                          # /*!
            ((?:[^!]|![^*]|!\*[^/])*)    #   (doc-string) -> captured
          !\*/\s+                        # !*/
          service\s+{name}               # service NAME
"""


class IdentType(Enum):
    """
    Use these enum entries with `get_doc_str()`.
    """
    STRUCT = 0
    SERVICE = 1
    RPC = 2


"""
Caches glob.glob() results for *.zs file searches in get_doc_str().
The dictionary points from a package path to amalgamated zserio code
for that package.
"""
zs_pkg_cache: Dict[str, str] = {}


def get_amalgamated_zs(pkg_path):
    global zs_pkg_cache
    if pkg_path in zs_pkg_cache:
        return zs_pkg_cache[pkg_path]
    zs_files = glob.glob(os.path.join(pkg_path, "**/*.zs"), recursive="True")
    result = ""
    for zs_file_path in zs_files:
        with open(zs_file_path) as zs_file:
            result += zs_file.read() + "\n"
    zs_pkg_cache[pkg_path] = result
    return result


def get_doc_str(*, ident_type: IdentType, pkg_path: str, ident: str, fallback: List[str] = None) -> List[str]:
    f"""
    Get a docstring for a particular zserio identifier. This method searches all .zs-files
    under `pkg_path` for a specific pattern given by `ident_type` and `ident`

    The following patterns are looked for:

      With `ident_type` IdentType.STRUCT:
        With ident as "path.to.package.NAME":
         {STRUCT_PATTERN}

      With `ident_type` IdentType.SERVICE)
        Same as STRUCT, except looking for "service NAME".

      With `ident_type` IdentType.RPC)
        With ident as "path.to.service.SERVICE.NAME":
         {RPC_DOC_PATTERN}

    The list of all capture group values is returned.
    """
    if fallback is None:
        fallback = []
    if not pkg_path:
        return fallback
    zs_src = get_amalgamated_zs(pkg_path)
    ident_parts = ident.split(".")
    pattern_format_replacements = {}
    if ident_type == IdentType.STRUCT:
        if not ident_parts:
            print("[ERROR] Need at least one identifier part to find struct docs.")
            return fallback
        pattern = STRUCT_PATTERN
        pattern_format_replacements["name"] = ident_parts[-1]
    elif ident_type == IdentType.SERVICE:
        if not ident_parts:
            print("[ERROR] Need at least one identifier part to find service docs.")
            return fallback
        pattern = SERVICE_PATTERN
        pattern_format_replacements["name"] = ident_parts[-1]
    elif ident_type == IdentType.RPC:
        if not ident_parts or len(ident_parts) < 2:
            print("[ERROR] Need at least tow identifiers (service.rpc-name) to find RPC docs.")
            return fallback
        pattern = RPC_DOC_PATTERN
        pattern_format_replacements["service_name"] = ident_parts[-2]
        pattern_format_replacements["rpc_name"] = ident_parts[-1]
    else:
        print("[ERROR] get_doc_str: Unsupported identifier type!")
        return fallback
    compiled_pattern = re.compile(pattern.format(**pattern_format_replacements), re.X)
    match = compiled_pattern.search(zs_src)
    if match:
        return list(match.groups())
    else:
        return fallback


def md_filter_definition(md: str) -> str:
    return re.sub(r"\n*\*\*[Dd]efinitions?[:\s]*\*\*\n*", "", md.strip()).strip()



import zserio
from os.path import dirname, abspath

working_dir = dirname(abspath(__file__))
zserio.generate(
    zs_dir=working_dir,
    main_zs_file="calculator.zs",
    gen_dir=working_dir,
    extra_args=["-withTypeInfoCode"])



import inspect
import base64
import zserio
import struct
import functools
from enum import Enum
from typing import Type, Tuple, Any, Dict, Union, Optional, List, get_type_hints, Iterator
from pyzswagcl import OAMethod, OAParam, OAParamFormat, ZSERIO_REQUEST_PART_WHOLE
from re import compile as re
from zserio.typeinfo import TypeInfo, MemberInfo, TypeAttribute, MemberAttribute

NONE_T = type(None)
SCALAR_T = (bool, int, float, str)
TYPE_INFO_CACHE = {}


class NotInstantiableReason(RuntimeError):
    def __init__(self, member_name: str, member_type: str, method_name: str = ""):
        super(NotInstantiableReason, self).__init__("\n" + inspect.cleandoc(f"""
        WARNING: Generating method `{method_name or 'unknown'}`:
        ----------------------------{"-"*len(method_name or 'unknown')}--
        
            The request must be passed as a blob, since the
            member `{member_name}` has non-flattable type
            `{member_type}`.
            
            Please add the following argument to zswag.gen,
            and remove any default parameter specifiers which
            might affect the configuration of `method_name`:
                
                -c {method_name}:blob
        """))
        self.member_name = member_name
        self.member_type = member_type


# Table here: https://docs.python.org/3/library/struct.html#format-characters
C_STRUCT_LITERAL_PER_TYPE_AND_SIZE: Dict[Tuple[Any, int], str] = {
    (bool, 1): "!b",
    (bool, 2): "!h",
    (bool, 4): "!i",
    (bool, 8): "!q",
    (int, 1): "!b",
    (int, 2): "!h",
    (int, 4): "!i",
    (int, 8): "!q",
    (float, 4): "!f",
    (float, 8): "!d",
}


# Recursive setattr function, adopted from SO:
#  https://stackoverflow.com/questions/31174295/getattr-and-setattr-on-nested-subobjects-chained-properties
def rsetattr(obj, attr, val):
    pre, _, post = attr.rpartition('.')
    return setattr(rgetattr(obj, pre) if pre else obj, post, val)


# Recursive getattr function, adopted from SO:
#  https://stackoverflow.com/questions/31174295/getattr-and-setattr-on-nested-subobjects-chained-properties
def rgetattr(obj, attr):
    def _getattr(obj, attr):
        if not hasattr(obj, attr):
            raise RuntimeError(
                f"\nERROR: `{attr}` does not exist not in {obj.__name__}; choices are ..." +
                "".join("\n  * "+choice for choice in dir(obj) if not choice.startswith("__")))
        return getattr(obj, attr)
    return functools.reduce(_getattr, [obj] + attr.split('.'))


# Get the request type for a zserio service method.
def service_method_request_type(service_instance: Any, method_name: str) -> Any:
    zserio_impl_function = getattr(service_instance, f"_{to_snake(method_name)}_impl")
    result = get_type_hints(zserio_impl_function)["request"]
    assert inspect.isclass(result)
    return result


# Adopted from zserio PythonSymbolConverter
def to_snake(s: str, patterns=(re("([a-z])([A-Z])"), re("([0-9A-Z])([A-Z][a-z])"))):
    for p in patterns:
        s = p.sub(r"\1_\2", s)
    return s.lower()


# Returns true if t is atomic (not a compound struct),
# false otherwise.
def is_scalar(t: TypeInfo):
    return t.py_type in SCALAR_T or issubclass(t.py_type, Enum)


# Retrieve zserio type info for a schema class -
# result is cached for maximum performance.
def cached_type_info(zserio_t) -> Optional[TypeInfo]:
    if not hasattr(zserio_t, "type_info"):
        return None
    key = f"{zserio_t.__module__}.{zserio_t.__qualname__}"
    if key not in TYPE_INFO_CACHE:
        TYPE_INFO_CACHE[key] = zserio_t.type_info()
    return TYPE_INFO_CACHE[key]


# Recursively find a member info (e.g. "myField1.myField2") in
# a TypeInfo. Member names may be pythonic. The input field path
# must be with original (not necessarily pythonic) schema field
# names. The first output tuple entry is the pythonic conversion
# ("my_field_1.my_field_2").
def find_field(t: TypeInfo, field: str, offset: int = 0, py_path="") -> Tuple[str, Optional[MemberInfo]]:
    next_dot = field.find(".", offset)
    subfield = field[offset:next_dot] if next_dot >= 0 else field[offset:]
    for _, member_info in members(t):
        if member_info.schema_name == subfield:
            if offset > 0:
                py_path += "."
            py_path += member_info.attributes[MemberAttribute.PROPERTY_NAME]
            if next_dot < 0:
                return py_path, member_info
            return find_field(member_info.type_info, field, next_dot+1, py_path)
    return "", None


# Simply yields all MemberInfo for a TypeInfo.
def members(t: TypeInfo, recursive=False, field_prefix="") -> Iterator[MemberInfo]:
    if TypeAttribute.FIELDS in t.attributes:
        for member in t.attributes[TypeAttribute.FIELDS]:
            field_name = field_prefix+"." if field_prefix else ""
            field_name += member.schema_name
            yield field_name, member
            if recursive:
                yield from members(member.type_info, recursive, field_name)


# Checks whether is a type is not trivially constructable:
# It (or a subtype) might take extra parameters, or be
# an array of non-scalar types.
def check_uninstantiable(t: TypeInfo, field_name="", recursive=False) -> Optional[NotInstantiableReason]:
    if TypeAttribute.PARAMETERS in t.attributes and t.attributes[TypeAttribute.PARAMETERS]:
        return NotInstantiableReason(field_name, t.schema_name)
    member_info: MemberInfo
    for _, member_info in members(t):
        subfield_name = f"{field_name}.{member_info.schema_name}" if field_name else member_info.schema_name
        if MemberAttribute.ARRAY_LENGTH in member_info.attributes and not is_scalar(member_info.type_info):
            return NotInstantiableReason(subfield_name, t.schema_name+"[]")
        if recursive:
            if reason := check_uninstantiable(member_info.type_info, subfield_name, recursive):
                return reason


# Returns a fully recursively initialized class instance
# of a zserio schema type, if possible.
def instantiate(t: Type) -> Any:
    result_instance = t()
    type_info = cached_type_info(t)
    if reason := check_uninstantiable(type_info):
        raise reason
    for _, member_info in members(type_info):
        field_name: str = member_info.attributes[MemberAttribute.PROPERTY_NAME]
        member_type_info = member_info.type_info
        if not is_scalar(member_type_info):
            compound_field_value = instantiate(member_type_info.py_type)
            setattr(result_instance, f"_{field_name}_", compound_field_value)
    return result_instance


# Get a byte buffer from a string which is encoded in a given format
def str_to_bytes(s: str, fmt: OAParamFormat) -> bytes:
    if fmt == OAParamFormat.BASE64:
        return base64.b64decode(s)
    elif fmt == OAParamFormat.BASE64URL:
        return base64.urlsafe_b64decode(s)
    elif fmt == OAParamFormat.HEX:
        return bytes.fromhex(s)
    else:  # if fmt in (OAParamFormat.BINARY, OAParamFormat.STRING):
        return bytes(s, encoding="raw_unicode_escape")


# Convert a single passed parameter value to it's correct type
def parse_param_value(param: OAParam, target_type: Type, value: str) -> Any:
    # Check if it's an enum ...
    if issubclass(target_type, Enum):
        return target_type(parse_param_value(param, int, value))
    # Check if the parameter format is native string conversion
    if param.format == OAParamFormat.STRING:
        if target_type is bool:
            return bool(int(value))
        return target_type(value)
    # Is it hex-to-int conversion?
    if param.format == OAParamFormat.HEX and target_type is int:
        return int(value, 16)
    # It is some kind of blob conversion!
    value_as_bytes = str_to_bytes(value, param.format)
    # Is it a serialized C type (char/uchar/float/double/long...)?
    struct_literal_key = (target_type, len(value_as_bytes))
    if struct_literal_key in C_STRUCT_LITERAL_PER_TYPE_AND_SIZE:
        return target_type(struct.unpack(
            C_STRUCT_LITERAL_PER_TYPE_AND_SIZE[struct_literal_key],
            value_as_bytes)[0])
    # Is it a string blob?
    if target_type is str:
        return str(value_as_bytes, 'utf-8')
    raise RuntimeError(f"Cannot convert {len(value_as_bytes)} to {target_type}")


# Convert an array of passed parameter values to their correct type
def parse_param_values(param: OAParam, target_type: Type, value: List[str]) -> List[Any]:
    return [parse_param_value(param, target_type, item) for item in value]


# Get a blob for a zserio request type, a set of request parameter values
# and an OpenAPI method path spec.
def request_object_blob(*, req_t: Type, headers: Dict[str, Any], spec: OAMethod, **kwargs) -> bytes:
    # Lazy instantiation of request object and type info
    req: Optional[req_t] = None
    req_t_info = cached_type_info(req_t)
    # Apply parameters
    param_name: str
    param: OAParam
    for param_name, param in spec.parameters.items():
        # Get raw string value
        value: Union[str, List[str]] = param.default_value
        param_name = param_name.replace("-", "_")
        if param_name in kwargs:
            value = kwargs[param_name]
        else:
            if param_name in headers:
                value = headers[param_name]
        # Convert string value to whole blob
        if param.field == ZSERIO_REQUEST_PART_WHOLE:
            return str_to_bytes(value, param.format)
        # First non-whole request parameter: Synthetic request object
        if not req:
            req = instantiate(req_t)
        # Convert string value to correct type
        pythonic_field_path, target_type = find_field(req_t_info, param.field)
        if not target_type:
            raise RuntimeError(f"Could not find field {param.field}!")
        if MemberAttribute.ARRAY_LENGTH in target_type.attributes:
            converted_value = parse_param_values(param, target_type.type_info.py_type, value)
        else:
            converted_value = parse_param_value(param, target_type.type_info.py_type, value)
        # Apply value to request object field
        rsetattr(req, pythonic_field_path, converted_value)
    # Serialise request object. Assert at least one parameter given.
    assert req
    serializer = zserio.BitStreamWriter()
    req.write(serializer)
    return serializer.byte_array



import connexion
import os
import inspect
import zserio
import sys
import yaml
from typing import Type
from flask import request as flask_request

from pyzswagcl import \
    parse_openapi_config, \
    OAMethod

from .reflect import request_object_blob, service_method_request_type, to_snake

# Name of variable that is added to controller
# to hold the service instance.
CONTROLLER_SERVICE_INSTANCE = "_service"

# Name of OpenApi extension field from which connexion
# reads the target wsgi function.
CONTROLLER_OPENAPI_FIELD = "x-openapi-router-controller"

# Utility function for slash conversion in format strings
def to_slashes(s: str):
    return s.replace("\\", "/")


# Raised if the controller passed to OAServer is missing a function
class IncompleteSchemaError(RuntimeError):
    def __init__(self, schema_path: str, fn_name: str):
        super(IncompleteSchemaError, self).__init__(f"Missing operation `{fn_name}` in {schema_path}!")
        self.schema_path = schema_path
        self.fn_name = fn_name


class OAServer(connexion.App):

    def __init__(self, *,
                 controller_module,
                 service_type: Type[zserio.ServiceInterface],
                 zs_pkg_path: str = None,
                 yaml_path: str = None):
        """
        Brief

            Marry a user-written app controller with a zserio-generated app server class
            (argument parser/response serialiser) and a fitting Swagger OpenApi spec.

            The OpenApi spec is auto-generated if the user does not specify an existing file.
            If the user specifies an empty YAML path, the yaml file is placed next to the
            zserio python-service source-file.

            If you have installed `pip install connexion[swagger-ui]`, you can view
            API docs of your service under [/prefix]/ui.

            Documentation for the service is automatically extracted if `zs_pkg_path` is issued.

        Code example

            In file my.app.__init__:

                from zserio_gen.my.service import Service
                from zswag import Server
                import

                app = Server(my.app.controller, Service)

            In file my.app.controller:

                # NOTE: Injected by Server, invisible to developer!
                #  In swagger yaml, the path specs reference `my.app.controller._service.myApi`
                _service = Service()
                _service.myApi = lambda request: _service._myApiMethod(request)
                _service._myApiImpl = my.app.controller.my_api

                # Written by user
                def my_api(request):
                    return "response"

        General call structure:

            OpenAPI `yaml_file` "paths/service" references
             Zserio Server instance injected method (ns.service.service(base64): blob), which calls
             ns.service._serviceMethod(blob), which calls
             ns.service._serviceImpl(value) which is remapped to
             User function (ns.serviceImpl(value): value).

        """
        if not yaml_path:
            service_module = sys.modules[service_type.__module__]
            yaml_path = os.path.join(
                os.path.dirname(os.path.abspath(service_module.__file__)),
                f"{service_module.__name__.split('.')[-1]}.{service_type.__name__}.yaml")
            print(f"Using yaml path {yaml_path}")
        self.yaml_path = yaml_path
        yaml_parent_path = os.path.dirname(yaml_path)
        self.zs_pkg_path = zs_pkg_path

        # Initialise zserio service
        self.service_type = service_type
        assert inspect.isclass(self.service_type)

        # Initialise zserio service working instance
        self.controller_path = controller_module.__name__
        self.controller = controller_module
        self.service_instance_path = self.controller_path+f".{CONTROLLER_SERVICE_INSTANCE}"
        if not hasattr(self.controller, CONTROLLER_SERVICE_INSTANCE):
            setattr(self.controller, CONTROLLER_SERVICE_INSTANCE, self.service_type())
        self.service_instance = getattr(self.controller, CONTROLLER_SERVICE_INSTANCE)

        # Verify or generate yaml file
        if not os.path.isfile(yaml_path):
            print("\n" + inspect.cleandoc(f"""
                ERROR: File does not exist: {yaml_path}
                ----------------------------{"-"*len(yaml_path)}
                
                You can generate the file by running ...
                    
                    python -m zswag.gen \\
                        --service {self.service_instance.SERVICE_FULL_NAME} \\
                        --input "{
                            to_slashes(os.path.abspath(os.path.dirname(os.path.dirname(
                                __import__(self.service_type.__module__.split('.')[0]).__file__
                            ))))
                        }" \\
                        --config post,body \\
                        --output "{to_slashes(yaml_path)}"
                        
                    The --config argument is a comma-separated list of tags
                    to specify OpenAPI options. For more info, please run
                    
                        python -m zswag.gen --help
            """))
            exit(1)

        self.spec = parse_openapi_config(yaml_path)
        self.verify_openapi_schema()

        # Re-route service impl methods
        for method_name in self.service_instance.method_names:
            method_snake_name = to_snake(method_name)
            user_function = getattr(self.controller, method_snake_name)
            zserio_modem_function = getattr(self.service_instance, f"_{method_snake_name}_method")
            request_type = service_method_request_type(self.service_instance, method_name)
            assert zserio_modem_function

            if not user_function or not inspect.isfunction(user_function):
                print(f"WARNING: The controller {self.controller_path} does not implement {method_snake_name}!")
                continue

            print(f"Found {self.controller_path}.{method_snake_name}.")
            if len(inspect.signature(user_function).parameters) != 1:
                print(f"ERROR: {self.controller_path}.{method_snake_name} must have single 'request' parameter!")
                continue

            method_spec: OAMethod = self.spec[method_name]

            def wsgi_method(fun=zserio_modem_function, spec=method_spec, req_t=request_type, **kwargs):
                if spec.body_request_object:
                    request_blob = kwargs["body"]
                else:
                    request_blob = request_object_blob(
                        req_t=req_t,
                        spec=spec,
                        headers=flask_request.headers,
                        **kwargs)
                try:
                    return bytes(fun(request_blob, None).byte_array)
                except zserio.PythonRuntimeException as e:
                    if str(e).startswith("BitStreamReader"):
                        return "Error in BitStreamReader: Could not parse malformed request.", 400
                    else:
                        return f"Internal Server Error: {e}", 500
            setattr(self.service_instance, method_name, wsgi_method)

            def method_impl(request, ctx=None, fun=user_function):
                return fun(request)
            setattr(self.service_instance, f"_{method_snake_name}_impl", method_impl)

        # Load spec and inject openapi-router-controller
        print(f"Loading spec from {yaml_path} ...")
        with open(yaml_path, 'r') as swagger_file:
            openapi = yaml.load(swagger_file, Loader=yaml.FullLoader)
        for path_name, path_spec in openapi["paths"].items():
            for meth_name, method_spec in path_spec.items():
                if CONTROLLER_OPENAPI_FIELD not in method_spec:
                    method_spec[CONTROLLER_OPENAPI_FIELD] = self.service_instance_path
                else:
                    print(f"{meth_name} {path_name}: Using pre-set {CONTROLLER_OPENAPI_FIELD}.")

        # Initialise connexion app
        super(OAServer, self).__init__(
            self.controller_path,
            specification_dir=yaml_parent_path)

        # Add the API according to the verified yaml spec.
        self.add_api(
            openapi,
            arguments={"title": f"REST API for {service_type.__name__}"},
            pythonic_params=False)

    def verify_openapi_schema(self):
        for method_name in self.service_instance.method_names:
            if method_name not in self.spec:
                raise IncompleteSchemaError(self.yaml_path, method_name)




import copy
import os
import random

import openapi_spec_validator.exceptions
import yaml
import sys
import inspect
import tempfile
import importlib
from typing import Optional, Dict, Tuple, List, IO
from argparse import ArgumentParser, FileType, RawTextHelpFormatter
from subprocess import CalledProcessError
import dataclasses as dc
from copy import deepcopy
from enum import Enum
from zserio.typeinfo import MemberAttribute
from openapi_spec_validator import validate_spec
import uuid
import zserio
from pyzswagcl import \
    ZSERIO_OBJECT_CONTENT_TYPE, \
    ZSERIO_REQUEST_PART_WHOLE, \
    ZSERIO_REQUEST_PART, \
    parse_openapi_config

from .reflect import \
    service_method_request_type, \
    rgetattr, \
    check_uninstantiable, \
    cached_type_info, \
    members as type_members, \
    find_field, \
    is_scalar
from .doc import get_doc_str, IdentType, md_filter_definition


class HttpParamLocation(Enum):
    QUERY = "query"
    BODY = "body"
    PATH = "path"
    HEADER = "header"


class HttpParamFormat(Enum):
    STRING = "string"
    BINARY = "binary"
    BYTE = "byte"
    BASE64 = "base64"
    BASE64URL = "base64url"
    HEX = "hex"


HTTP_METHOD_TAGS = ("get", "put", "post", "delete")
FLATTEN_TAG = "flat"
BLOB_TAG = "blob"
SECURITY_ASSIGNMENT_TAG = "security="
PATH_ASSIGNMENT_TAG = "path="
WILDCARD_CONFIG = "*"


def argdoc(s: str):
    return "\n"+inspect.cleandoc(s)+"\n\n"


def less_indent_formatter(prog):
    return RawTextHelpFormatter(prog, max_help_position=8, width=80)


@dc.dataclass
class ParamSpecifier:
    """"Can be used to annotate a method config with a desired
    location, format and parameter name for a request-part field."""
    request_part: str            # e.g. "request_member.subfield"
    name: Optional[str]          # e.g. "subfieldParam"
    location: HttpParamLocation  # e.g. QUERY
    format: HttpParamFormat        # e.g. BASE64
    style: Optional[str] = None
    explode: Optional[str] = None


@dc.dataclass
class MethodConfig:
    """User-specified set of parameters to control how a zserio service method
    is converted into an OpenAPI-file entry."""
    name: str
    http_method: str = "post"
    param_loc: Optional[HttpParamLocation] = None  # None -> Whole request blob in body
    flatten: bool = False
    param_specifiers: Optional[List[ParamSpecifier]] = dc.field(default_factory=list)
    security: Optional[str] = None
    path: Optional[str] = None
    openapi_docstring: str = ""
    openapi_return_type: str = ""
    openapi_arg_type: str = "Unknown"
    openapi_result_doc: str = ""
    openapi_arg_doc: str = ""
    openapi_parameters: dict = dc.field(default_factory=dict)


class OpenApiGenError(RuntimeError):
    """Raised by OpenApiSchemaGenerator when something goes wrong."""
    def __init__(self, what: str):
        super(OpenApiGenError, self).__init__(what)


class OpenApiSchemaGenerator:

    def __init__(self, *,
                 service: str,
                 path: str,
                 package: Optional[str] = None,
                 config: Optional[List[str]] = None,
                 output: IO,
                 base_config: Optional[IO],
                 zserio_src_root: Optional[str]):

        # Process service name and package path
        self.service_name = service
        self.zs_pkg_path = zserio_src_root
        service_name_parts = service.split(".") + ["Service"]

        if os.path.isdir(path):
            # Generate OpenAPI from existing Python code
            sys.path.append(path)
            python_module = importlib.import_module(f"{service_name_parts[0]}.api")
            self.zs_pkg_path = None
        else:
            if self.zs_pkg_path is None:
                self.zs_pkg_path = os.path.abspath(os.path.dirname(path))
                path = os.path.basename(path)
            full_path = os.path.join(self.zs_pkg_path, path)
            if not os.path.isfile(full_path):
                raise OpenApiGenError(f"The path '{full_path}' is neither a zserio file nor a pythonpath.")
            # Generate OpenAPI from zserio code. Must generate
            # intermediate Python source to inspect service.
            if package is None:
                package = f"zswag_gen_{uuid.uuid1().hex}"
            if service_name_parts[0] != package:
                service_name_parts = [package] + service_name_parts
            try:
                gen_dir = tempfile.mkdtemp("zswag.gen")
                python_module = zserio.generate(
                    zs_dir=self.zs_pkg_path,
                    main_zs_file=path,
                    top_level_package=package,
                    gen_dir=gen_dir,
                    extra_args=["-withTypeInfoCode"])
            except CalledProcessError as e:
                raise OpenApiGenError(f"Failed to parse zserio sources:\n{e.stderr}")
        if not python_module:
            raise OpenApiGenError(f"Could not import {service_name_parts[0]}.api!")
        self.service_type = rgetattr(python_module, ".".join(service_name_parts[1:]))
        self.service_instance = self.service_type()

        # Process method config tags ...
        self.config: Dict[str, MethodConfig] = dict()
        self.config[WILDCARD_CONFIG] = default_entry = MethodConfig(WILDCARD_CONFIG)
        self.output = output
        self.base_config = dict()
        # ... first load base-config.
        if base_config:
            self.base_config = yaml.load(base_config, yaml.Loader)
        if methods_base_config := self.base_config.get("methods", None):
            # If there are wildcard base-config tags, process them first.
            if default_config_tags := methods_base_config.get(WILDCARD_CONFIG, None):
                default_entry = self.add_method_config_from_tags(WILDCARD_CONFIG, default_config_tags, default_entry)
            # Then process tags for other methods from the base-config.
            for method_name, tags in ((m, t) for m, t in methods_base_config.items() if m != WILDCARD_CONFIG):
                self.add_method_config_from_tags(method_name, tags, default_entry)
        # ... then process additional tags from the command line.
        if config:
            for entry in config:
                method_name = WILDCARD_CONFIG
                parts = entry.split(":")
                if len(parts) > 1:
                    method_name = parts[0]
                tag_list = [tag.strip() for tag in parts[-1].split(",")]
                default_entry = self.add_method_config_from_tags(method_name, tag_list, default_entry)

    def add_method_config_from_tags(
            self,
            method_name: str,
            tag_list: List[str],
            default_entry: MethodConfig) -> MethodConfig:
        new_config = deepcopy(default_entry)
        new_config.name = method_name
        # Make sure that HTTP method tags are always processed first!
        tag_list.sort(key=lambda x: x in HTTP_METHOD_TAGS, reverse=True)
        for tag in tag_list:
            if tag in HTTP_METHOD_TAGS:
                new_config.http_method = tag
            elif any(tag == loc.value for loc in HttpParamLocation):
                new_config.param_loc = HttpParamLocation(tag)
                if new_config.param_loc == HttpParamLocation.BODY and new_config.http_method == "get":
                    raise OpenApiGenError(f"Cannot use `body` tag with HTTP GET.")
            elif tag == FLATTEN_TAG:
                new_config.flatten = True
            elif tag == BLOB_TAG:
                new_config.flatten = False
            elif tag.startswith(SECURITY_ASSIGNMENT_TAG):
                new_config.security = tag[len(SECURITY_ASSIGNMENT_TAG):]
            elif tag.startswith(PATH_ASSIGNMENT_TAG):
                if method_name == WILDCARD_CONFIG:
                    raise OpenApiGenError(f"Refusing to apply '{tag}' to ALL methods. Likely not on purpose.")
                else:
                    new_config.path = tag[len(PATH_ASSIGNMENT_TAG):]
            elif "?" in tag:
                try:
                    request_part, specifiers_part = tag.split("?")
                    spec = dict(specifier.split("=") for specifier in specifiers_part.split("&"))
                    par_loc = HttpParamLocation(spec.get("in"))
                    if par_loc == HttpParamLocation.BODY:
                        if new_config.http_method == "get":
                            raise OpenApiGenError("Cannot use `location=body` with HTTP GET.")
                        par_format = HttpParamFormat.BINARY
                        par_name = "body"
                    else:
                        par_name, par_format = spec.get("name"), HttpParamFormat(spec.get("format", "string"))
                    if any(par.name == par_name for par in new_config.param_specifiers):
                        raise OpenApiGenError(f"Encountered duplicate use of parameter name '{par_name}' in the same method!")
                    new_config.param_specifiers.append(ParamSpecifier(request_part, par_name, par_loc, par_format))
                    # Process style and explode entries
                    if style := spec.get("style", None):
                        new_config.param_specifiers[-1].style = style.lower()
                    if explode := spec.get("explode", None):
                        new_config.param_specifiers[-1].explode = explode.lower()
                except (ValueError, IndexError) as e:
                    raise OpenApiGenError(f"Encountered malformed parameter specifier tag '{tag}' ({e})!")
                except KeyError as missing_key:
                    raise OpenApiGenError(f"Parameter specifier tag {tag} has no value for '{missing_key}' key.")
            else:
                raise OpenApiGenError(f"Did not understand tag '{tag}'.")
        if method_name in self.config and method_name != WILDCARD_CONFIG:
            print(f"[WARNING] Overwriting config for method {method_name}!")
        self.config[method_name] = new_config
        if method_name == WILDCARD_CONFIG:
            default_entry = new_config
        return default_entry

    def config_for_method(self, method_name: str) -> MethodConfig:
        if method_name in self.config:
            return self.config[method_name]
        else:
            result = copy.deepcopy(self.config[WILDCARD_CONFIG])
            result.name = method_name
            return result

    def generate(self):
        service_name_parts = self.service_instance.service_full_name.split(".")
        schema = {
            "openapi": "3.0.0",
            "info": self.base_config.get("info", {
                "title": ".".join(service_name_parts[1:]),
                "description": md_filter_definition(get_doc_str(
                    ident_type=IdentType.SERVICE,
                    pkg_path=self.zs_pkg_path,
                    ident=self.service_instance.service_full_name,
                    fallback=[f"REST API for {'.'.join(service_name_parts[1:])} service."]
                )[0]),
                "contact": {
                    "email": "TODO@TODO.TODO"
                },
                "license": {
                    "name": "TODO"
                },
                "version": "TODO",
            }),
            "servers": self.base_config.get("servers", []),
            "paths": {
                method_info.path: {
                    method_info.http_method: {
                        "summary": method_info.openapi_docstring,
                        "description": method_info.openapi_docstring,
                        "operationId": method_info.name,
                        **method_info.openapi_parameters,
                        "responses": {
                            "200": {
                                "description": method_info.openapi_result_doc,
                                "content": {
                                    ZSERIO_OBJECT_CONTENT_TYPE: {
                                        "schema": {
                                            "type": "string",
                                            "format": "binary"
                                        }
                                    }
                                }
                            }
                        }
                    },
                } for method_info in (
                    self.process_method_config(method_name)
                    for method_name in self.service_instance.method_names)
            }
        }
        if security_schemes := self.base_config.get("securitySchemes", None):
            schema["components"] = {
                "securitySchemes": security_schemes
            }
        if security := self.base_config.get("security", None):
            schema["security"] = security
        print(f"[INFO] Validating with openapi-spec-validator ... ", end="")
        try:
            validate_spec(schema)
        except openapi_spec_validator.exceptions.OpenAPIValidationError as e:
            print()
            raise OpenApiGenError(f"Spec is invalid: {e}")
        print("OK")
        print(f"[INFO] Writing to '{self.output.name}' ... ", end="")
        yaml.dump(schema, self.output, default_flow_style=False)
        self.output.close()
        print("OK")
        if os.path.isfile(self.output.name):
            print(f"[INFO] Validating with zswag parser ... ", end="")
            try:
                parse_openapi_config(self.output.name)
            except RuntimeError as e:
                print()
                raise OpenApiGenError(f"Spec is invalid: {e}")
            print("OK")
        else:
            print(f"[INFO] Skipping zswag parser validation.")
        print(f"[INFO] Done.")

    def process_method_config(self, method_name: str) -> MethodConfig:
        result = self.config_for_method(method_name)
        req_t = service_method_request_type(self.service_instance, result.name)
        req_t_info = cached_type_info(req_t)
        assert req_t_info
        method_ident = f"{self.service_instance.service_full_name}.{method_name}"
        if not result.path:
            print(f"[INFO] Auto-generating path for `{method_ident}`.")
            result.path = f"/{method_name}"
        # Generate doc-strings
        if self.zs_pkg_path:
            doc_strings = get_doc_str(
                ident_type=IdentType.RPC,
                pkg_path=self.zs_pkg_path,
                ident=method_ident)
            if doc_strings:
                result.openapi_docstring = doc_strings[0]
                result.openapi_return_type = doc_strings[1]
                result.openapi_arg_type = doc_strings[2]
                result.openapi_result_doc = md_filter_definition(get_doc_str(
                    ident_type=IdentType.STRUCT,
                    pkg_path=self.zs_pkg_path,
                    ident=result.openapi_return_type,
                    fallback=[f"### struct {result.openapi_return_type}"])[0])
                result.openapi_arg_doc = md_filter_definition(get_doc_str(
                    ident_type=IdentType.STRUCT,
                    pkg_path=self.zs_pkg_path,
                    ident=result.openapi_arg_type,
                    fallback=[f"### struct {result.openapi_arg_type}"])[0])
        # Convert simple legacy instructions into parameter specifiers
        if not result.param_specifiers:
            if result.param_loc == HttpParamLocation.BODY or (not result.param_loc and not result.flatten):
                result.param_specifiers.append(
                    ParamSpecifier(ZSERIO_REQUEST_PART_WHOLE, "", HttpParamLocation.BODY, HttpParamFormat.BINARY))
            elif result.flatten:
                if not_instantiable_reason := check_uninstantiable(req_t_info):
                    raise OpenApiGenError(str(not_instantiable_reason))
                for field_name, member_info in type_members(req_t_info, True):
                    if is_scalar(member_info.type_info):
                        result.param_specifiers.append(
                            ParamSpecifier(
                                field_name, field_name.replace(".", "__"),
                                result.param_loc or HttpParamLocation.QUERY,
                                HttpParamFormat.STRING))
            elif result.param_loc:
                result.param_specifiers.append(
                    ParamSpecifier(
                        ZSERIO_REQUEST_PART_WHOLE, "requestBody", result.param_loc, HttpParamFormat.BINARY))
        # Convert MethodConfig.param_specifiers to MethodConfig.openapi_parameters
        self.process_method_parameters(result)
        return result

    def process_method_parameters(self, config: MethodConfig):
        req_t = service_method_request_type(self.service_instance, config.name)
        req_t_info = cached_type_info(req_t)
        # Convert parameter specifiers to parameters-JSON
        openapi_param_list = []
        for param_specifier in config.param_specifiers:
            # Easy - parameter only indicates binary transfer of the
            # request object in the body
            if param_specifier.location == HttpParamLocation.BODY:
                config.openapi_parameters["requestBody"] = {
                    "description": config.openapi_arg_doc,
                    "content": {
                        ZSERIO_OBJECT_CONTENT_TYPE: {
                            "schema": {
                                "type": "string"
                            }
                        }
                    }
                }
                continue
            # Fallthrough - the parameter has a name, a format and a request part
            # Process type-info first and determine if the field even exists.
            openapi_schema_info = {
                "type": "string",
                "format": param_specifier.format.name.lower(),
            }
            if param_specifier.request_part != ZSERIO_REQUEST_PART_WHOLE:
                _, member_info = find_field(req_t_info, param_specifier.request_part)
                if not member_info:
                    raise OpenApiGenError(f"Could not find field '{param_specifier.request_part}' in {req_t_info.schema_name}!")
                # If the member is an array, we must indicate this in OpenAPI
                if MemberAttribute.ARRAY_LENGTH in member_info.attributes:
                    openapi_schema_info.update({
                        "type": "array",
                        "items": {
                            "type": "string"
                        }
                    })
            # If the parameter is to be placed in the path, make sure there's a placeholder
            if param_specifier.location == HttpParamLocation.PATH:
                placeholder = f"{{{param_specifier.name}}}"
                if placeholder not in config.path:
                    print(f"[INFO] Appending /{placeholder} to path for {config.name}")
                    config.path += f"/{placeholder}"
            # Append the OpenAPI information for the new parameter
            openapi_param_list.append({
                "in": param_specifier.location.value.lower(),
                "name": param_specifier.name,
                "description": config.openapi_arg_doc,
                "required": True,
                ZSERIO_REQUEST_PART: param_specifier.request_part,
                **({"allowEmptyValue": True} if param_specifier.location == HttpParamLocation.QUERY else {}),
                "schema": openapi_schema_info
            })
            # Add style and explode hints
            if param_specifier.style:
                openapi_param_list[-1]["style"] = param_specifier.style
            if param_specifier.explode:
                openapi_param_list[-1]["explode"] = param_specifier.explode == "true"
        # Process security scheme
        if config.security is not None:
            if config.security == "":
                print(f"[WARNING] {config.name} has explicit empty security setting.")
                config.openapi_parameters["security"] = []
            else:
                config.openapi_parameters["security"] = [{config.security: []}]
        # Set the finalized parameter list
        config.openapi_parameters["parameters"] = openapi_param_list


if __name__ == "__main__":

    parser = ArgumentParser(
        "Zserio OpenApi YAML Generator",
        formatter_class=less_indent_formatter)
    parser.add_argument("-s", "--service", nargs=1, required=True,
                        metavar="service-identifier", help=argdoc("""
                        Fully qualified zserio service identifier.
                        
                        Example:
                            -s my.package.ServiceClass
                        """))
    parser.add_argument("-i", "--input", nargs=1, metavar="zserio-or-python-path",
                        required=True, help=argdoc("""
                        Can be either ...
                        (A) Path to a zserio .zs file. Must be either a top-
                            level entrypoint (e.g. all.zs), or a subpackage
                            (e.g. services/myservice.zs) in conjunction with
                            a "--zserio-source-root|-r <dir>" argument.
                        (B) Path to parent dir of a zserio Python package.
                        
                        Examples:
                            -i path/to/schema/main.zs         (A)
                            -i path/to/python/package/parent  (B) 
                        """))
    parser.add_argument("-r", "--zserio-source-root", nargs=1, metavar="zserio-src-root-dir",
                        required=False, help=argdoc("""
                        When -i specifies a zs file (Option A), indicate the
                        directory for the zserio -src directory argument. If
                        not specified, the parent directory of the zs file
                        will be used.
                        """))
    parser.add_argument("-p", "--package", nargs=1, metavar="top-level-package",
                        required=False, help=argdoc("""
                        When -i specifies a zs file (Option A), indicate
                        that a specific top-level zserio package name
                        should be used.
                        
                        Examples:
                            -p zserio_pkg_name
                        """))
    parser.add_argument("-c", "--config", nargs="+", metavar="tags",
                        action="append", help=argdoc("""
                        Configuration tags for a specific or all methods.
                        The argument syntax follows this pattern:
                            
                           [(service-method-name):](comma-separated-tags)
                            
                        Note: The -c argument may be applied multiple times.
                        The `comma-separated-tags` must be a list of tags
                        which indicate OpenApi method generator preferences.
                        The following tags are supported:
                        
                        get|put|post|delete : HTTP method tags
                                query|path| : Parameter location tags
                                header|body 
                                  flat|blob : Flatten request object,
                                              or pass it as whole blob.
                          (param-specifier) : Specify parameter name, format
                                              and location for a specific
                                              request-part. See below.
                            security=(name) : Set a particular security
                                              scheme to be used. The scheme
                                              details must be provided through
                                              the --base-config-yaml.
                         path=(method-path) : Set a particular method path.
                                              May contain placeholders for
                                              path params.
                                                   
                        A (param-specifier) tag has the following schema:
                        
                            (field?name=...
                                  &in=[path|body|query|header]
                                  &format=[binary|base64|hex]
                                  [&style=...]
                                  [&explode=...])
                        
                        Examples:
                        
                          Expose all methods as POST, but `getLayerByTileId`
                          as GET with flat path-parameters:
                          
                            `-c post getLayerByTileId:get,flat,path`
                            
                          For myMethod, put the whole request blob into the a
                          query "data" parameter as base64:
                          
                            `-c myMethod:*?name=data&in=query&format=base64`
                            
                          For myMethod, set the "AwesomeAuth" auth scheme:
                          
                            `-c myMethod:security=AwesomeAuth`
                            
                          For myMethod, provide the path and place myField
                          explicitely in a path placeholder:
                          
                            `-c 'myMethod:path=/my-method/{param},...
                                 myField?name=param&in=path&format=string'`
                            
                        Note:
                            * The HTTP-method defaults to `post`.
                            * The parameter 'in' defaults to `query` for
                              `get`, `body` otherwise.
                            * If a method uses a parameter specifier, the
                              `flat`, `body`, `query`, `path`, `header` and
                              `body`-tags are ignored.
                            * The `flat` tag is only meaningful in conjunction
                              with `query` or `path`.
                            * An unspecific tag list (no service-method-name)
                              affects the defaults only for following, not
                              preceding specialized tag assignments.
                        """))
    parser.add_argument("-o", "--output", nargs=1, type=FileType("w"), default=[sys.stdout],
                        metavar="output", help=argdoc("""
                        Output file path. If not specified, the output will be
                        written to stdout.
                        """))
    parser.add_argument("-b", "--base-config-yaml", nargs=1, type=FileType("r"), required=False, default=[None],
                        help=argdoc("""
                        Base configuration file. Can be used to fully or partially
                        substitute --config arguments, and to provide additional
                        OpenAPI information. The YAML file must look like this:
                        
                          method: # Optional method tags dictionary
                            <method-name|*>: <list of config tags>
                          securitySchemes: ... # Optional OpenAPI securitySchemes
                          info: ...            # Optional OpenAPI info section
                          servers: ...         # Optional OpenAPI servers section
                          security: ...        # Optional OpenAPI global security
                        """))

    args = parser.parse_args(sys.argv[1:])
    try:
        OpenApiSchemaGenerator(
            service=args.service[0],
            path=args.input[0] if args.input else None,
            package=args.package[0] if args.package else None,
            config=[arg for args in args.config for arg in args] if args.config else [],
            output=args.output[0],
            base_config=args.base_config_yaml[0] if args.base_config_yaml else None,
            zserio_src_root=args.zserio_source_root[0] if args.zserio_source_root else None).generate()
    except OpenApiGenError as e:
        print(f"[ERROR] {e}")
        exit(1)



import setuptools
import sys
import codecs
import os

# Read first argument as version, pass rest on to setuptools
VERSION = sys.argv[1]
del sys.argv[1]

# Determine the directory containing this setup.py file
setup_dir = os.path.dirname(os.path.abspath(__file__))
readme_path = os.path.join(setup_dir, "README.md")

with codecs.open(readme_path, "r", "utf-8") as fh:
    long_description = fh.read()

required_url = []
required = [f"pyzswagcl=={VERSION}"]
requirements_path = os.path.join(setup_dir, "requirements.txt")
with open(requirements_path, "r") as freq:
    for line in freq.read().split():
        if "://" in line:
            required_url.append(line)
        else:
            required.append(line)

setuptools.setup(
    name="zswag",
    version=VERSION,
    url="https://github.com/klebert-engineering/zswag",
    author="Klebert Engineering",
    author_email="team@klebert-engineering.de",

    description="OpenAPI service transport layer for zserio.",
    long_description=long_description,
    long_description_content_type="text/markdown",

    package_dir={'': 'libs'},
    packages=['zswag', 'zswag.test.calc'],
    include_package_data=True,
    package_data={
        "zswag": [
            "test/calc/api.yaml",
            "test/calc/calculator.zs"
        ]
    },

    install_requires=required,
    dependency_links=required_url,
    python_requires='>=3.10',

    license="BSD-3 Clause",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: BSD License"
     ],
)
